;;;;
;;;
;;; Prctica de Introduccin a la Inteligencia Artifical
;;; Curso: 3 ETSII-ULL
;;; Profesor: Patricio Garca
;;; Titulo: Sistema de Produccin para el problema de 
;;;         Trasvase de agua entre cubos
;;; Fecha: Abril,1998
;;; Necesario: Librera "shell.lsp" para creacin de Sistemas
;;;            de produccin (LOAD "shell")
;;;            Librera "q_search.lsp" para bsquedas genricas
;;;            (LOAD "q_search")
;;;
;;; Uso:
#|
;;;       Cargar primero:

(load "shell")
(load "q_search")
(load "cubos")

;;;       Para bsqueda en anchura:

(ba)

;;;       Para bsqueda en profundidad:

(bp)

;;;       Ojo, se cicla por no tener tratamiento de estados visitados
|#
;;;       
;;; Modificaciones:
;;;    Mayo 2004: 
;;;       Cambios debidos a modificacin de q_search para 
;;;       llamar a generic_search con una lista con el estado
;;;       inicial.
;;;       Cambios para que IMP-EST devuelva el estado que imprime
;;;       Cambios debidos a modificacin de shell.lsp para llamar 
;;;       a shell como paquete
;;;    Mayo 2006:
;;;       Cambio de texto maysculas/minsculas.
;;;
;;; Pendiente por hacer:
;;;       Incluir devolucin de camino
;;;       Tratamiento de estados visitados
;;;

; Usa libreras para reglas y para busquedas

(use-package "SHELL")
(use-package "Q_SEARCH")


; Constante versin

(defconstant *version-cubos* "may-2006"
  "Versin del programas cubos.")


; Estructura para estados
; -----------------------

(defstruct estado
  "Estructura para estados del problema de los cubos."
  (cubo1 0)     ; Cantidad de litros de agua del cubo1
  (cubo2 0))    ; Cantidad de litros de agua del cubo2


; Variables globales 
; ------------------

;Capacidades de los cubos

(defvar capacidad1 8 
  "Capacidad por defecto del cubo1.")

(defvar capacidad2 6 
  "Capacidad por defecto del cubo2.")


;Litros finales deseados en uno de los cubos

(defvar litrosfinales 4
  "Litros finales a obtener en el estado objetivo.")


; Definicin de lista de estados iniciales

(defvar estados-iniciales (list (make-estado))
  "Lista de estados iniciales.")


;Definicion de condicin de finalizacin de bsqueda
; --------------------------------------------------

(defun estados-objetivo-p (estado)
  "Chequea si ha de detenerse la bsqueda en el problema de los cubos."
  (if estado
      (or (= (estado-cubo1 estado) litrosfinales)
	  (= (estado-cubo2 estado) litrosfinales))
      t))


; Definicin de funciones para las reglas
; ---------------------------------------

;Regla1: Llena cubo 1

(defun p1 (est)
  "Precondicin regal 1."
  (/= (estado-cubo1 est) capacidad1))

(defun a1 (est)
  "Accin regla 1."
  (make-estado :cubo2 (estado-cubo2 est)
	       :cubo1 capacidad1))


;Regla2: Llena cubo 2

(defun p2  (est)
  "Precondicin regal 2."
  (/= (estado-cubo2 est) capacidad2))

(defun a2 (est)
  "Accin regla 2."
  (make-estado :cubo1 (estado-cubo1 est)
	       :cubo2 capacidad2))

;Regla3: Vacia cubo 1
(defun p3  (est)
  "Precondicin regal 3."
  (/= (estado-cubo1 est) 0))

(defun a3 (est)
  "Accin regla 3."
  (make-estado :cubo2 (estado-cubo2 est)
		 :cubo1 0))


;Regla4: Vacia cubo 2
(defun p4  (est)
  "Precondicin regal 4."
  (/= (estado-cubo2 est) 0))

(defun a4 (est)
  "Accin regla 4."
  (make-estado :cubo1 (estado-cubo1 est)
	       :cubo2 0))


;Regla5: Trasvasa del 1 al 2
(defun p5  (est)
  "Precondicin regal 5."
  (and (/= (estado-cubo1 est) 0)
       (/= (estado-cubo2 est) capacidad2)))

(defun a5 (est)
  "Accin regla 5."
  (let ((cant (min (estado-cubo1 est) 
		   (- capacidad2 
		      (estado-cubo2 est)))))
    (make-estado :cubo1 (- (estado-cubo1 est) cant)
		 :cubo2 (+ (estado-cubo2 est) cant))))


;Regla6: Trasvasa del 2 al 1
(defun p6  (est)
  "Precondicin regal 6."
  (and (/= (estado-cubo1 est) capacidad1)
       (/= (estado-cubo2 est) 0)))

(defun a6 (est)
  "Accin regla 6."
  (let ((cant (min (estado-cubo2 est) 
		   (- capacidad1 
		      (estado-cubo1 est)))))
    (make-estado :cubo1 (+ (estado-cubo1 est) cant)
		 :cubo2 (- (estado-cubo2 est) cant))))


; Insercin las reglas en el sistema de produccin
; ------------------------------------------------

(regla #'p1 #'a1)
(regla #'p2 #'a2)
(regla #'p3 #'a3)
(regla #'p4 #'a4)
(regla #'p5 #'a5)
(regla #'p6 #'a6)


; Funcin para imprimir un estado
; -------------------------------

(defun imp-est (est)
  "Imprime un estado del problema de los cubos."
  (if est
      (progn
	(format t "[~d ~d] " (estado-cubo1 est) (estado-cubo2 est))
	est)
    nil))



;Llamada a bsqueda en profundidad
; --------------------------------

(defun bp ()
  "Bsqueda en profundidad del problema de los cubos."
  (imp-est 
   (generic-search estados-iniciales
		   #'estados-objetivo-p
		   #'hijos
		   :display-fn #'imp-est
		   :merge-fn #'(lambda (new-states)
				 (nconc new-states *search-queue*))
		   :dequeue-fn #'(lambda () (pop *search-queue*)))))


;Llamada a bsqueda en anchura
; ----------------------------

(defun ba ()
  "Bsqueda en anchura del problema de los cubos."
  (imp-est 
   (generic-search estados-iniciales
		   #'estados-objetivo-p
		   #'hijos
		   :display-fn #'imp-est
		   :merge-fn #'(lambda (new-states)
				 (nconc *search-queue* new-states))
		   :dequeue-fn #'(lambda () (pop *search-queue*)))))

